#include <string.h>

#ifdef FOR_NSL

#include "lt-memory.h"
#include "nsllib.h"

#define Malloc salloc
#define Realloc srealloc
#define Free sfree

#else

#include "system.h"

#endif

#include "charset.h"
#include "string16.h"
#include "dtd.h"

char8 *url_merge(const char8 *url, const char8 *base,
		char8 **_scheme, char8 **_host, int *_port, char8 **_path);

const char8 *DefaultTypeName[DT_enum_count] = {
    "#required",
    "bogus1",
    "#implied",
    "bogus2"
    "none",
    "#fixed",
};

const char8 *ContentTypeName[CT_enum_count] = {
    "mixed",
    "any",
    "bogus1",
    "bogus2",
    "empty",
    "element"
};

const char8 *AttributeTypeName[AT_enum_count] = {
    "cdata",
    "bogus1",
    "bogus2",
    "nmtoken",
    "bogus3",
    "entity",
    "idref",
    "bogus4",
    "bogus5",
    "nmtokens",
    "bogus6",
    "entities",
    "idrefs",
    "id",
    "notation",
    "enumeration"
};

static Char *Strndup(const Char *old, int len)
{
    Char *new = Malloc((len+1) * sizeof(Char));

    if(!new)
	return 0;

    memcpy(new, old, len * sizeof(Char));
    new[len] = 0;

    return new;
}

/* DTDs */

Dtd NewDtd(void)
{
    Dtd d;

    if(!(d = Malloc(sizeof(*d))))
	return 0;

    d->name = 0;
    d->entities = 0;
    d->parameter_entities = 0;
    d->predefined_entities = 0;
#ifdef FOR_NSL
    d->doctype = 0;
#else
    d->elements = 0;
    d->notations = 0;
#endif

    return d;
}

/* Entities */

Entity NewExternalEntityN(const Char *name, int namelen, const char8 *publicid,
			 const char8 *systemid, NotationDefinition notation,
			 Entity parent)
{
    Entity e;

    if(!(e = Malloc(sizeof(*e))))
	return 0;
    if(name)
	if(!(name = Strndup(name, namelen)))
	    return 0;

    e->type = ET_external;
    e->name = name;
    e->base_url = 0;
    e->encoding = CE_unknown;
    e->next = 0;
    e->parent = parent;

    e->publicid = publicid;
    e->systemid = systemid;
    e->notation = notation;

    e->version_decl = 0;
    e->encoding_decl = 0;
    e->standalone_decl = 0;

    e->url = 0;

    return (Entity)e;
}

Entity NewInternalEntityN(const Char *name, int namelen, const Char *text, Entity parent,
			 int line_offset, int line1_char_offset, 
			 int matches_parent_text)
{
    Entity e;

    if(!(e = Malloc(sizeof(*e))))
	return 0;
    if(name)
	if(!(name = Strndup(name, namelen)))
	    return 0;

    e->type = ET_internal;
    e->name = name;
    e->base_url = 0;
    e->encoding = InternalCharacterEncoding;
    e->next = 0;
    e->parent = parent;

    e->text = text;
    e->line_offset = line_offset;
    e->line1_char_offset = line1_char_offset;
    e->matches_parent_text = matches_parent_text;

    e->url = 0;

    return (Entity)e;
}

const char8 *EntityURL(Entity e)
{
    /* Have we already determined the URL? */

    if(e->url)
	return e->url;

    if(e->type == ET_internal)
    {
	if(e->parent)
	{
	    const char8 *url = EntityURL(e->parent);
	    if(url)
		e->url = strdup8(url);
	}
    }
    else
	e->url = url_merge(e->systemid, 
			   e->parent ? EntityBaseURL(e->parent) : 0, 
			   0, 0, 0, 0);

    return e->url;
}

/* Returns the URL of the entity if it has one, otherwise the system ID.
   It will certainly have a URL if it was successfully opened by EntityOpen.
   Intended for error messages, so never returns NULL. */

const char8 *EntityDescription(Entity e)
{
    if(e->url)
	return e->url;

    if(e->type == ET_external)
	return e->systemid;

    if(e->parent)
	return EntityDescription(e->parent);

    return "<unknown>";
}

void EntitySetBaseURL(Entity e, const char8 *url)
{
    e->base_url = strdup8(url);
}

const char8 *EntityBaseURL(Entity e)
{
    if(e->base_url)
	return e->base_url;

    if(e->type == ET_internal)
    {
	if(e->parent)
	    return EntityBaseURL(e->parent);
	else
	    return 0;
    }

    return EntityURL(e);
}

Entity DefineEntity(Dtd dtd, Entity e, int pe)
{
    if(pe)
    {
	e->next = dtd->parameter_entities;
	dtd->parameter_entities = e;
    }
    else
    {
	e->next = dtd->entities;
	dtd->entities = e;
    }

    return e;
}

Entity FindEntityN(Dtd dtd, const Char *name, int namelen, int pe)
{
    Entity e;

    if(!pe)
	for(e = dtd->predefined_entities; e; e = e->next)
	    if(Strncmp(name, e->name, namelen) == 0 && e->name[namelen] == 0)
		return e;


    for(e = pe ? dtd->parameter_entities : dtd->entities; e; e=e->next)
	if(Strncmp(name, e->name, namelen) == 0 && e->name[namelen] == 0)
	    return e;

    return 0;
}

/* Elements */

ElementDefinition DefineElementN(Dtd dtd, const Char *name, int namelen, 
				 ContentType type, Char *content)
{
#ifdef FOR_NSL
    static struct element_definition e;
    RHTEntry *entry;
    struct element_content *c;

    /* XXX what if this fails? */
    entry = DeclareElement(dtd->doctype, name, namelen, 0, (ctVals)type);

    e.doctype = dtd->doctype;
    e.name = (Char *)dtd->doctype->elements+entry->keyptr;
    e.elsum = (NSL_ElementSummary_I *)(dtd->doctype->elementBase+entry->eval);

    if(type == CT_element)
    {
	if(!(c = Malloc(sizeof(*c))))
	    return 0;
	c->elsum = e.elsum;
	c->content = content;
	c->next = e.doctype->element_content;
	e.doctype->element_content = c;
    }

    return &e;
#else
    ElementDefinition e;

    if(!(e = Malloc(sizeof(*e))) || !(name = Strndup(name, namelen)))
	return 0;

    e->tentative = 0;
    e->name = name;
    e->namelen = namelen;
    e->type = type;
    e->content = content;
    e->attributes = 0;
    e->next = dtd->elements;
    dtd->elements = e;

    return e;
#endif
}

ElementDefinition TentativelyDefineElementN(Dtd dtd, 
					    const Char *name, int namelen)
{
#ifdef FOR_NSL
    static struct element_definition e;
    RHTEntry *entry;

    /* XXX what if this fails? */
    entry = DeclareElement(dtd->doctype, name, namelen, 0, (ctVals)CT_any);
    e.doctype = dtd->doctype;
    e.name = (Char *)dtd->doctype->elements+entry->keyptr;
    e.elsum = (NSL_ElementSummary_I *)(dtd->doctype->elementBase+entry->eval);
    /* XXX use the omitStart field to note that it's tentative. */
    e.elsum->omitStart |= 2;
    return &e;
#else
    ElementDefinition e;

    if(!(e = Malloc(sizeof(*e))) || !(name = Strndup(name, namelen)))
	return 0;

    e->tentative = 1;
    e->name = name;
    e->namelen = namelen;
    e->type = CT_any;
    e->content = 0;
    e->attributes = 0;
    e->next = dtd->elements;
    dtd->elements = e;

    return e;
#endif
}

ElementDefinition RedefineElement(ElementDefinition e, ContentType type,
				  Char *content)
{
#ifdef FOR_NSL
    struct element_content *c;

    e->elsum->contentType = type;
    e->elsum->omitStart &= ~2;
    if(type == CT_element)
    {
	if(!(c = Malloc(sizeof(*c))))
	    return 0;
	c->elsum = e->elsum;
	c->content = content;
	c->next = e->doctype->element_content;
	e->doctype->element_content = c;
    }
#else
    e->tentative = 0;
    e->type = type;
    e->content = content;
#endif
    return e;
}

ElementDefinition FindElementN(Dtd dtd, const Char *name, int namelen)
{
#ifdef FOR_NSL
    /* Derived from FindElementAndName */

    static struct element_definition e;
    RHTEntry *entry;
    NSL_ElementSummary_I *elsum;

    if(!dtd->doctype)
	return 0;

    entry = rsearch(name, namelen, dtd->doctype->elements);
    if(!entry)
	return 0;

    elsum = (NSL_ElementSummary_I *)(dtd->doctype->elementBase+entry->eval);
    while(elsum->numAttr < 0)
	elsum = (NSL_ElementSummary_I *)(-elsum->numAttr);

    e.doctype = dtd->doctype;
    e.name = (Char *)dtd->doctype->elements+entry->keyptr;
#if 0
    /* We don't use this so don't waste time on it XXX */
    e.namelen = Strlen(e.name);
#endif
    e.elsum = elsum;
    e.tentative = ((e.elsum->omitStart & 2) != 0);

    return &e;
#else
    ElementDefinition e, *p;

    for(p = &dtd->elements, e = *p; e; p = &e->next, e = *p)
	if(namelen == e->namelen && *name == *e->name &&
	   memcmp(name, e->name, namelen*sizeof(Char)) == 0)
	{
	    *p = e->next;
	    e->next = dtd->elements;
	    dtd->elements = e;
	    return e;
	}

    return 0;
#endif
}

/* Attributes */

AttributeDefinition
    DefineAttributeN(ElementDefinition element, const Char *name, int namelen,
		     AttributeType type, Char **allowed_values,
		     DefaultType default_type, const Char *default_value)
{
#ifdef FOR_NSL
    int nav = 0;

    NSL_Doctype_I *doctype = element->doctype;

    if(!doctype)
	return 0;

    if(allowed_values)
	for(; *allowed_values; allowed_values++)
	    nav++;

    /* XXX what if this fails? */
    name = DeclareAttr(doctype, name, namelen,
		       (NSL_Attr_Dec_Value)type,
		       allowed_values ? allowed_values[0] : 0, nav,
		       (NSL_ADefType)default_type, default_value,
		       &element->elsum, element->name);

    return (AttributeDefinition)FindAttrSpec(element->elsum, name);
#else
    AttributeDefinition a;

    if(!(a= Malloc(sizeof(*a))) || !(name = Strndup(name, namelen)))
	return 0;

    a->name = name;
    a->namelen = namelen;
    a->type = type;
    a->allowed_values = allowed_values;
    a->default_type = default_type;
    a->default_value = default_value;
    a->next = element->attributes;
    element->attributes = a;

    return a;
#endif
}

AttributeDefinition FindAttributeN(ElementDefinition element,
				   const Char *name, int namelen)
{
#ifdef FOR_NSL
    if(!element->doctype)
	return 0;

    if(!(name = AttrUniqueName(element->doctype, name, namelen)))
	return 0;
    return (AttributeDefinition)FindAttrSpec(element->elsum, name);
#else
    AttributeDefinition a;

    for(a = element->attributes; a; a = a->next)
	if(namelen == a->namelen &&
	   memcmp(name, a->name, namelen * sizeof(Char)) == 0)
	    return a;

    return 0;
#endif
}

AttributeDefinition NextAttributeDefinition(ElementDefinition element,
					    AttributeDefinition previous)
{
#ifdef FOR_NSL
    return 0;			/* not implemented */
#else
    if(previous)
	return previous->next;
    else
	return element->attributes;
#endif
}

/* Notations */

NotationDefinition DefineNotationN(Dtd dtd, const Char *name, int namelen, 
				  const char8 *publicid, const char8 *systemid)
{
    NotationDefinition n;

    if(!(n = Malloc(sizeof(*n))) || !(name = Strndup(name, namelen)))
	return 0;

    n->name = name;
    n->tentative = 1;
    n->systemid = systemid;
    n->publicid = publicid;
    n->next = dtd->notations;
    dtd->notations = n;

    return n;
}

NotationDefinition TentativelyDefineNotationN(Dtd dtd,
					      const Char *name, int namelen)
{
    NotationDefinition n;

    if(!(n = Malloc(sizeof(*n))) || !(name = Strndup(name, namelen)))
	return 0;

    n->name = name;
    n->tentative = 1;
    n->systemid = 0;
    n->publicid = 0;
    n->next = dtd->notations;
    dtd->notations = n;

    return n;
}

NotationDefinition RedefineNotation(NotationDefinition n,
				  const char8 *publicid, const char8 *systemid)
{
    n->tentative = 0;
    n->systemid = systemid;
    n->publicid = publicid;

    return n;
}

NotationDefinition FindNotationN(Dtd dtd, const Char *name, int namelen)
{
    NotationDefinition n;

    for(n = dtd->notations; n; n = n->next)
	if(Strncmp(name, n->name, namelen) == 0 && n->name[namelen] == 0)
	    return n;

    return 0;
}


