#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#ifdef FOR_NSL

#include "lt-memory.h"
#include "nsllib.h"

#define Malloc salloc
#define Realloc srealloc
#define Free sfree

#else

#include "system.h"

#endif

#include "charset.h"
#include "string16.h"
#include "dtd.h"
#include "input.h"
#include "ctype16.h"

static int FileReadBlock(void **handle, unsigned char *buf, int max_count);
static int FileSeek(void **handle, int offset);
static void FileClose(void *handle);

#ifdef USE_ZLIB
#include "zlib.h"
#ifdef macintosh
#include <fcntl.h>
#include <unix.h>
#endif
static int GzipReadBlock(void **handle, unsigned char *buf, int max_count);
static int GzipSeek(void **handle, int offset);
static void GzipClose(void *handle);
#endif

static int StringReadBlock(void **handle, unsigned char *buf, int max_count);
static int StringSeek(void **handle, int offset);

static int get_translated_line1(InputSource s);

FILE *url_open(const char8 *url, const char8 *base, const char8 *type,
	       char8 **merged_url);
char8 *default_base_url(void);

InputSource SourceFromStream(const char8 *description, FILE *file)
{
    Entity e;

    e = NewExternalEntity(0, 0, description, 0, 0);
    if(!strchr8(description, '/'))
	EntitySetBaseURL(e, default_base_url());

    return NewInputSource(e, FileReadBlock, FileSeek, 0, (void *)file);
}

InputSource EntityOpen(Entity e)
{
    if(e->type == ET_external)
    {
	FILE *file;
	const char8 *url = EntityURL(e);
#ifdef USE_ZLIB
	int len;
#endif

	if(!url || !(file = url_open(url, 0, "r", 0)))
	    return 0;

#ifdef USE_ZLIB
	/*
	 * here we need to do something in the case that the URL
	 * corresponds to something compressed. 
	 */
	len = strlen(url);
	if(url[len-3] == '.' && url[len-2] == 'g' && url[len-1]== 'z') 
	{	
		
		 gzFile gfile;
		

#ifdef macintosh
	/* It appears that duplicating the file
	 * descriptor and calling  fclose is a 
	 * bad thing to do on the Mac. So we don't do it...
	 */
		gfile=gzdopen(fileno(file),"rb");
#else 	
		gfile = gzdopen(dup(fileno(file)),"r");
		fclose(file);
#endif
		
		return NewInputSource(e,GzipReadBlock,GzipSeek,GzipClose,(void*) gfile);
	}
#endif
	return NewInputSource(e, FileReadBlock, FileSeek, FileClose,
			      (void *)file);

	
    }
    else
	return NewInputSource(e, StringReadBlock, StringSeek, 0,
			      (void *)e->text);
}

static int FileReadBlock(void **handle, unsigned char *buf, int max_count)
{
    FILE *file = *handle;
    int count = 0;

    count = fread(buf, 1, max_count, file);

    return ferror(file) ? -1 : count;
}

static int FileSeek(void **handle, int offset)
{
    FILE *file = *handle;

    return fseek(file, offset, SEEK_SET);
}

static void FileClose(void *handle)
{
    FILE *f = handle;

    fclose(f);
}


#ifdef USE_ZLIB

static int GzipReadBlock(void **handle, unsigned char *buf, int max_count)
{
    gzFile file = (gzFile) *handle;
    int count = 0;
    int gzerr;
    const char * errorString;

    count = gzread(file,buf,max_count);
  
	errorString = gzerror(file,&gzerr);
	if(gzerr != 0 && gzerr != Z_STREAM_END)
		return -1;
		
	return count;
}

static int GzipSeek(void ** handle,int offset) {
	return -1;
}

static void GzipClose(void * handle) {
	gzclose((gzFile) handle);
	}

#endif

/* XXX change this!!! */
static int StringReadBlock(void **handle, unsigned char *buf, int max_count)
{
    Char *p = *handle;
    int c, count = 0;

    max_count /= sizeof(Char);

    while(count < max_count && (c = *p))
    {
	p++;
	((Char *)buf)[count++] = c;
    }

    *handle = p;

    return count * sizeof(Char);
}

static int StringSeek(void **handle, int offset)
{
#if 1
    return -1;			/* until we fix reference to e below! */
#else
    Char *p = *handle;

    *handle = e->text + offset;

    return 0;			/* What if it's off the end? XXX */
#endif
}

InputSource NewInputSource(Entity e, ReadProc read, SeekProc seek, 
			   CloseProc close, void *handle)
{
    InputSource source;

    if(!(source = Malloc(sizeof(*source))))
	return 0;

    source->line = 0;
    source->line_alloc = 0;
    source->line_length = 0;
    source->next = 0;
    source->seen_eoe = 0;

    source->entity = e;

    source->read = read;
    source->seek = seek;
    source->close = close;
    source->handle = handle;

    source->bytes_consumed = 0;
    source->bytes_before_current_line = 0;
    source->line_end_was_cr = 0;
    source->line_number = 0;
    source->not_read_yet = 1;

    source->nextin = source->insize = 0;

    source->parent = 0;

    return source;
}

int SourceLineAndChar(InputSource s, int *linenum, int *charnum)
{
    Entity e = s->entity, f = e->parent;

    if(e->type == ET_external)
    {
	*linenum = s->line_number;
	*charnum = s->next;
	return 1;
    }

    if(f && f->type == ET_external)
    {
	if(e->matches_parent_text)
	{
	    *linenum = e->line_offset + s->line_number;
	    *charnum = (s->line_number == 0 ? e->line1_char_offset : 0) +
		       s->next;
	    return 1;
	}
	else
	{
	    *linenum = e->line_offset;
	    *charnum = e->line1_char_offset;
	    return 0;
	}
    }

    if(f && f->matches_parent_text)
    {
	*linenum = f->line_offset + e->line_offset;
	*charnum = (e->line_offset == 0 ? f->line1_char_offset : 0) +
	    e->line1_char_offset;
	return 0;
    }

    return -1;
}

void SourcePosition(InputSource s, Entity *entity, int *byte_offset)
{
    *entity = s->entity;
    *byte_offset = SourceTell(s);
}

int SourceTell(InputSource s)
{
#if CHAR_SIZE == 8
    return s->bytes_before_current_line + s->next;
#else
    switch(s->entity->encoding)
    {
    case CE_ISO_10646_UCS_2B:
    case CE_UTF_16B:
    case CE_ISO_10646_UCS_2L:
    case CE_UTF_16L:
	return s->bytes_before_current_line + 2 * s->next;
    case CE_ISO_8859_1:
	return s->bytes_before_current_line + s->next;
    case CE_UTF_8:
	if(s->complicated_utf8_line)
	{
	    /* examine earlier chars in line to see how many bytes they used */
	    int i, c, n=0;
	    for(i = 0; i < s->next; i++)
	    {
		c = s->line[i];
		if(c <= 0x7f)
		    n += 1;
		else if(c <= 0x7ff)
		    n += 2;
		else if(c >= 0xd800 && c <= 0xdfff)
		    /* One of a surrogate pair, count 2 each */
		    n += 2;
		else if(c <= 0xffff)
		    n += 3;
		else if(c <= 0x1ffff)
		    n += 4;
		else if(c <= 0x3ffffff)
		    n += 5;
		else
		    n += 6;

	    }
	    return s->bytes_before_current_line + n;
	}
	else
	    return s->bytes_before_current_line + s->next;
    default:
	return -1;
    }
#endif
}

int SourceSeek(InputSource s, int offset)
{
    s->line_length = 0;
    s->next = 0;
    s->seen_eoe = 0;
    s->bytes_consumed = s->bytes_before_current_line = offset;
    s->nextin = s->insize = 0;
    /* XXX line number will be wrong! */
    s->line_number = -999999;
    return s->seek(&s->handle, offset);
}

static int get_translated_line(InputSource s)
{
    Entity e = s->entity;
    Char *p;

    /* This is a hack, pending some reorganisation */

    if(e->type == ET_external)
	return get_translated_line1(s);

    if(!*(Char *)s->handle)
    {
	s->line_length = 0;
	return 0;
    }

    s->line = s->handle;
    for(p=s->line; *p && *p != '\n'; p++)
	;
    if(*p)
	p++;
    s->handle = p;
    s->line_length = p - s->line;

    s->bytes_before_current_line = (char *)s->handle - (char *)e->text;

    return 0;
}

static int get_translated_line1(InputSource s)
{
    unsigned int c;		/* can't use Char, it might be >0x10000 */
    unsigned char *inbuf = s->inbuf;
    int nextin = s->nextin, insize = s->insize;
    int startin = s->nextin;
    Char *outbuf = s->line;
    int outsize = s->line_alloc;
    int nextout = 0;
    int remaining = 0;
    int ignore_linefeed = s->line_end_was_cr;

#if CHAR_SIZE == 16
    CharacterEncoding enc = s->entity->encoding;
    int more, i;
    s->complicated_utf8_line = 0;
#endif

    s->line_end_was_cr = 0;
    s->bytes_before_current_line = s->bytes_consumed;

    while(1)
    {
	/* There are never more characters than bytes in the input */
	if(outsize < nextout + (insize - nextin))
	{
	    outsize = nextout + (insize - nextin);
	    outbuf = Realloc(outbuf, outsize * sizeof(Char));
	}

	while(nextin < insize)
	{
#if CHAR_SIZE == 8
	    c = inbuf[nextin++];
#else
	    switch(enc)
	    {
	    case CE_ISO_10646_UCS_2B:
	    case CE_UTF_16B:
		if(nextin+2 > insize)
		    goto more_bytes;
		c = (inbuf[nextin] << 8) + inbuf[nextin+1];
		nextin += 2;
		break;
	    case CE_ISO_10646_UCS_2L:
	    case CE_UTF_16L:
		if(nextin+2 > insize)
		    goto more_bytes;
		c = (inbuf[nextin+1] << 8) + inbuf[nextin];
		nextin += 2;
		break;
	    case CE_ISO_8859_1:
		c = inbuf[nextin++];
		break;
	    case CE_UTF_8:
		c = inbuf[nextin++];
		if(c <= 0x7f)
		    break;
		if(c <= 0xc0 || c >= 0xfe)
		{
		    fprintf(stderr, "Illegal UTF-8 start byte <0x%x>\n", c);
		    return -1;
		}
		if(c <= 0xdf)
		{
		    c &= 0x1f;
		    more = 1;
		}
		else if(c <= 0xef)
		{
		    c &= 0x0f;
		    more = 2;
		}
		else if(c <= 0xf7)
		{
		    c &= 0x07;
		    more = 3;
		}
		else if(c <= 0xfb)
		{
		    c &= 0x03;
		    more = 4;
		}
		else
		{
		    c &= 0x01;
		    more = 5;
		}
		if(nextin+more > insize)
		{
		    nextin--;
		    goto more_bytes;
		}
		s->complicated_utf8_line = 1;
		for(i=0; i<more; i++)
		    c = (c << 6) + (inbuf[nextin++] & 0x3f);
		break;
	    default:
		fprintf(stderr,
			"read from entity with unsupported encoding!\n");
		return -1;
	    }
#endif
	    if(c > 0x110000 || (c < 0x10000 && !is_xml_legal(c)))
		if(!(enc == CE_UTF_16L || enc == CE_UTF_16B) ||
		   c < 0xd800 || c > 0xdfff)
		    /* We treat the surrogates as legal because we didn't
		       combine them when translating from UTF-16.  XXX */
		{
		    fprintf(stderr, "Error: illegal character <0x%x>\n", c);
		    return -1;
		}

	    if(c == '\n' && ignore_linefeed)
	    {
		/* Ignore lf at start of line if last line ended with cr */
		ignore_linefeed = 0;
		s->bytes_before_current_line += (nextin - startin);
	    }		
	    else
	    {
		ignore_linefeed = 0;
		if(c == '\r')
		{
		    s->line_end_was_cr = 1;
		    c = '\n';
		}

#if CHAR_SIZE == 16
		if(c >= 0x10000)
		{
		    /* Use surrogates */
		    outbuf[nextout++] = ((c - 0x10000) >> 10) + 0xd800;
		    outbuf[nextout++] = ((c - 0x10000) & 0x3ff) + 0xdc00;
		}
		else
		    outbuf[nextout++] = c;
#else
		outbuf[nextout++] = c;
#endif

		if(c == '\n')
		{
		    s->nextin = nextin;
		    s->insize = insize;
		    s->bytes_consumed += (nextin - startin);
		    s->line = outbuf;
		    s->line_alloc = outsize;
		    s->line_length = nextout;
		    return 0;
		}
	    }
	}

#if CHAR_SIZE == 16
    more_bytes:
	/* Copy down any partial character */

	remaining = insize - nextin;
	for(i=0; i<remaining; i++)
	    inbuf[i] = inbuf[nextin + i];
#endif

	/* Get another block */

	s->bytes_consumed += (nextin - startin);

	insize = s->read(&s->handle, 
			 inbuf+insize-nextin, sizeof(s->inbuf)-remaining);
	nextin = startin = 0;

	if(insize <= 0)
	{
		s->nextin = nextin;
		s->insize = 0;
		s->line = outbuf;
		s->line_alloc = outsize;
		s->line_length = nextout;
		return insize;
	}

	insize += remaining;
    }
}

void determine_character_encoding(InputSource s)
{
    Entity e = s->entity;
    int nread;
    unsigned char *b = (unsigned char *)s->inbuf;

    b[0] = b[1] = b[2] = b[3] = 0;

    while(s->insize < 4)
    {
	nread = s->read(&s->handle, s->inbuf + s->insize, 4 - s->insize);
	if(nread == -1)
	    return;
	if(nread == 0)
	    break;
	s->insize += nread;
    }

#if 0
    if(b[0] == 0 && b[1] == 0 && b[2] == 0 && b[3] == '<')
	e->encoding = CE_ISO_10646_UCS_4B;
    else if(b[0] == '<' && b[1] == 0 && b[2] == 0 && b[3] == 0)
	e->encoding = CE_ISO_10646_UCS_4L;
    else
#endif
    if(b[0] == 0xfe && b[1] == 0xff)
    {
	e->encoding = CE_UTF_16B;
	s->nextin = 2;
    }
    else if(b[0] == 0 && b[1] == '<' && b[2] == 0 && b[3] == '?')
	e->encoding = CE_UTF_16B;
    else if(b[0] == 0xff && b[1] == 0xfe)
    {
	e->encoding = CE_UTF_16L;
	s->nextin = 2;
    }
    else if(b[0] == '<' && b[1] == 0 && b[2] == '?' && b[3] == 0)
	e->encoding = CE_UTF_16L;
    else
    {
#if CHAR_SIZE == 8	
	e->encoding = CE_ISO_8859_1;
#else
        e->encoding = CE_UTF_8;
#endif
    }
}

int get_with_fill(InputSource s)
{
    assert(!s->seen_eoe);

    if(get_translated_line(s) != 0)
    {
	/* It would be nice to pass this up to the parser, but we don't
	   know anything about parsers here! */
	fprintf(stderr, "I/O error on stream <%s>\n",
		EntityDescription(s->entity));

	/* Restore old line and return EOE (is this the best thing to do?) */
	s->line_length = s->next;
	s->seen_eoe = 1;
	return XEOE;
    }

    if(s->line_length == 0)
    {
	/* Restore old line */
	s->line_length = s->next;
	s->seen_eoe = 1;
	return XEOE;
    }

    s->next = 0;

    if(s->not_read_yet)
	s->not_read_yet = 0;
    else
	s->line_number++;

    return s->line[s->next++];
}
